# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations  # needed for the TYPE_CHECKING import

import re
from functools import partial
from typing import (
    TYPE_CHECKING,
    Optional,
    Union,
    Literal,
    Sequence,
    Dict,
    Callable,
    List,
    Any,
    Tuple,
)

from evdev.ecodes import EV_KEY, EV_REL, EV_ABS
from gi.repository import Gtk

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import (
    MappingData,
    UIMapping,
    MappingType,
)
from inputremapper.configs.paths import PathUtils
from inputremapper.configs.validation_errors import (
    pydantify,
    MissingMacroOrKeyError,
    MacroButTypeOrCodeSetError,
    SymbolAndCodeMismatchError,
    MissingOutputAxisError,
    WrongMappingTypeForKeyError,
    OutputSymbolVariantError,
)
from inputremapper.exceptions import DataManagementError
from inputremapper.gui.components.output_type_names import OutputTypeNames
from inputremapper.gui.data_manager import DataManager, DEFAULT_PRESET_NAME
from inputremapper.gui.gettext import _
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
    MessageType,
)
from inputremapper.gui.messages.message_data import (
    PresetData,
    StatusData,
    CombinationRecorded,
    UserConfirmRequest,
    DoStackSwitch,
)
from inputremapper.gui.utils import CTX_APPLY, CTX_ERROR, CTX_WARNING, CTX_MAPPING
from inputremapper.injection.injector import (
    InjectorState,
    InjectorStateMessage,
)
from inputremapper.logging.logger import logger

if TYPE_CHECKING:
    # avoids gtk import error in tests
    from inputremapper.gui.user_interface import UserInterface


MAPPING_DEFAULTS = {"target_uinput": "keyboard"}


class Controller:
    """Implements the behaviour of the gui."""

    def __init__(
        self,
        message_broker: MessageBroker,
        data_manager: DataManager,
    ) -> None:
        self.message_broker = message_broker
        self.data_manager = data_manager
        self.gui: Optional[UserInterface] = None

        self.button_left_warn = False
        self._attach_to_events()

    def set_gui(self, gui: UserInterface):
        """Let the Controller know about the user interface singleton.."""
        self.gui = gui

    def _attach_to_events(self) -> None:
        self.message_broker.subscribe(MessageType.groups, self._on_groups_changed)
        self.message_broker.subscribe(MessageType.preset, self._on_preset_changed)
        self.message_broker.subscribe(MessageType.init, self._on_init)
        self.message_broker.subscribe(
            MessageType.preset, self._publish_mapping_errors_as_status_msg
        )
        self.message_broker.subscribe(
            MessageType.mapping, self._publish_mapping_errors_as_status_msg
        )

    def _on_init(self, __):
        """Initialize the gui and the data_manager."""
        # make sure we get a groups_changed event when everything is ready
        # this might not be necessary if the reader-service takes longer to provide the
        # initial groups
        self.data_manager.publish_groups()
        self.data_manager.publish_uinputs()

    def _on_groups_changed(self, _):
        """Load the newest group as soon as everyone got notified
        about the updated groups."""

        if self.data_manager.active_group is not None:
            # don't jump to a different group and preset suddenly, if the user
            # is already looking at one
            logger.debug("A group is already active")
            return

        group_key = self.get_a_group()
        if group_key is None:
            logger.debug("Could not find a group")
            return

        self.load_group(group_key)

    def _on_preset_changed(self, data: PresetData):
        """Load a mapping as soon as everyone got notified about the new preset."""
        if data.mappings:
            mappings = list(data.mappings)
            mappings.sort(
                key=lambda mapping: (
                    mapping.format_name() or mapping.input_combination.beautify()
                )
            )
            combination = mappings[0].input_combination
            self.load_mapping(combination)
            self.load_input_config(combination[0])
        else:
            # send an empty mapping to make sure the ui is reset to default values
            self.message_broker.publish(MappingData(**MAPPING_DEFAULTS))

    def _on_combination_recorded(self, data: CombinationRecorded):
        combination = self._auto_use_as_analog(data.combination)
        self.update_combination(combination)

    def _format_status_bar_validation_errors(self) -> Optional[Tuple[str, str]]:
        if not self.data_manager.active_preset:
            return None

        if self.data_manager.active_preset.is_valid():
            self.message_broker.publish(StatusData(CTX_MAPPING))
            return None

        mappings = list(self.data_manager.active_preset)

        # Move the selected (active) mapping to the front, so that it is checked first.
        active_mapping = self.data_manager.active_mapping
        if active_mapping is not None:
            mappings.remove(active_mapping)
            mappings.insert(0, active_mapping)

        for mapping in mappings:
            if not mapping.has_input_defined():
                # Empty mapping, nothing recorded yet so nothing can be configured,
                # therefore there isn't anything to validate.
                continue

            position = mapping.format_name()
            error_strings = self._get_ui_error_strings(mapping)

            if len(error_strings) == 0:
                continue

            if len(error_strings) > 1:
                msg = _('%d Mapping errors at "%s", hover for info') % (
                    len(error_strings),
                    position,
                )
                tooltip = "– " + "\n– ".join(error_strings)
            else:
                msg = f'"{position}": {error_strings[0]}'
                tooltip = error_strings[0]

            return msg.replace("\n", " "), tooltip

        return None

    def _publish_mapping_errors_as_status_msg(self, *__) -> None:
        """Send mapping ValidationErrors to the MessageBroker."""
        validation_result = self._format_status_bar_validation_errors()

        if validation_result is None:
            return

        self.show_status(
            CTX_MAPPING,
            validation_result[0],
            validation_result[1],
        )

    @staticmethod
    def format_error_message(mapping, error_type, error_message: str) -> str:
        """Check all the different error messages which are not useful for the user."""
        # There is no more elegant way of comparing error_type with the base class.
        # https://github.com/pydantic/pydantic/discussions/5112
        if (
            pydantify(MacroButTypeOrCodeSetError) in error_type
            or pydantify(SymbolAndCodeMismatchError) in error_type
        ) and mapping.input_combination.defines_analog_input:
            return _(
                "Remove the macro or key from the macro input field "
                "when specifying an analog output"
            )

        if (
            pydantify(MacroButTypeOrCodeSetError) in error_type
            or pydantify(SymbolAndCodeMismatchError) in error_type
        ) and not mapping.input_combination.defines_analog_input:
            return _(
                "Remove the Analog Output Axis when specifying a macro or key output"
            )

        if pydantify(MissingOutputAxisError) in error_type:
            error_message = _(
                "The input specifies an analog axis, but no output axis is selected."
            )
            if mapping.output_symbol is not None:
                event = [
                    event
                    for event in mapping.input_combination
                    if event.defines_analog_input
                ][0]
                error_message += _(
                    "\nIf you mean to create a key or macro mapping "
                    "go to the advanced input configuration"
                    ' and set a "Trigger Threshold" for '
                    f'"{event.description()}"'
                )
            return error_message

        if pydantify(WrongMappingTypeForKeyError) in error_type:
            error_message = _(
                "The input specifies a key, but the output type is not "
                f'"{OutputTypeNames.key_or_macro}".'
            )

            if mapping.output_type in (EV_ABS, EV_REL):
                error_message += _(
                    "\nIf you mean to create an analog axis mapping go to the "
                    'advanced input configuration and set an input to "Use as Analog".'
                )

            return error_message

        if pydantify(MissingMacroOrKeyError) in error_type:
            return _("Missing macro or key")

        return error_message

    @staticmethod
    def _get_ui_error_strings(mapping: UIMapping) -> List[str]:
        """Get a human readable error message from a mapping error."""
        validation_error = mapping.get_error()

        if validation_error is None:
            return []

        formatted_errors = []

        for error in validation_error.errors():
            if pydantify(OutputSymbolVariantError) in error["type"]:
                # this is rather internal, when this error appears in the gui, there is
                # also always another more readable error at the same time that explains
                # this problem.
                continue

            error_string = f'"{mapping.format_name()}": '
            error_message = error["msg"]
            error_location = error["loc"][0]
            if error_location != "__root__":
                error_string += f"{error_location}: "

            # check all the different error messages which are not useful for the user
            formatted_errors.append(
                Controller.format_error_message(
                    mapping,
                    error["type"],
                    error_message,
                )
            )

        return formatted_errors

    def get_a_preset(self) -> str:
        """Attempts to get the newest preset in the current group
        creates a new preset if that fails."""
        try:
            return self.data_manager.get_newest_preset_name()
        except FileNotFoundError:
            pass
        self.data_manager.create_preset(self.data_manager.get_available_preset_name())
        return self.data_manager.get_newest_preset_name()

    def get_a_group(self) -> Optional[str]:
        """Attempts to get the group with the newest preset
        returns any if that fails."""
        try:
            return self.data_manager.get_newest_group_key()
        except FileNotFoundError:
            pass

        keys = self.data_manager.get_group_keys()
        return keys[0] if keys else None

    def copy_preset(self):
        """Create a copy of the active preset and name it `preset_name copy`."""
        name = self.data_manager.active_preset.name
        match = re.search(r" copy *\d*$", name)
        if match:
            name = name[: match.start()]

        self.data_manager.copy_preset(
            self.data_manager.get_available_preset_name(f"{name} copy")
        )
        self.message_broker.publish(DoStackSwitch(1))

    def _auto_use_as_analog(self, combination: InputCombination) -> InputCombination:
        """If output is analog, set the first fitting input to analog."""
        if self.data_manager.active_mapping is None:
            return combination

        if not self.data_manager.active_mapping.is_analog_output():
            return combination

        if combination.find_analog_input_config():
            # something is already set to do that
            return combination

        for i, input_config in enumerate(combination):
            # find the first analog input and set it to "use as analog"
            if input_config.type in (EV_ABS, EV_REL):
                logger.info("Using %s as analog input", input_config)

                # combinations and input_configs are immutable, a new combination
                # is created to fit the needs instead
                combination_list = list(combination)
                combination_list[i] = input_config.modify(analog_threshold=0)
                new_combination = InputCombination(combination_list)
                return new_combination

        return combination

    def update_combination(self, combination: InputCombination):
        """Update the input_combination of the active mapping."""
        combination = self._auto_use_as_analog(combination)

        try:
            self.data_manager.update_mapping(input_combination=combination)
            self.save()
        except KeyError:
            self.show_status(
                CTX_MAPPING,
                f'"{combination.beautify()}" already mapped to something else',
            )
            return

        if combination.is_problematic():
            self.show_status(
                CTX_WARNING,
                _("ctrl, alt and shift may not combine properly"),
                _(
                    "Your system might reinterpret combinations with those after they "
                    + "are injected, and by doing so break them. Play around with the "
                    + 'advanced "Release Input" toggle.'
                ),
            )

    def move_input_config_in_combination(
        self,
        input_config: InputConfig,
        direction: Union[Literal["up"], Literal["down"]],
    ):
        """Move the active_input_config up or down in the input_combination of the
        active_mapping."""
        if (
            not self.data_manager.active_mapping
            or len(self.data_manager.active_mapping.input_combination) == 1
        ):
            return
        combination: Sequence[InputConfig] = (
            self.data_manager.active_mapping.input_combination
        )

        i = combination.index(input_config)
        if (
            i + 1 == len(combination)
            and direction == "down"
            or i == 0
            and direction == "up"
        ):
            return

        if direction == "up":
            combination = (
                list(combination[: i - 1])
                + [input_config]
                + [combination[i - 1]]
                + list(combination[i + 1 :])
            )
        elif direction == "down":
            combination = (
                list(combination[:i])
                + [combination[i + 1]]
                + [input_config]
                + list(combination[i + 2 :])
            )
        else:
            raise ValueError(f"unknown direction: {direction}")
        self.update_combination(InputCombination(combination))
        self.load_input_config(input_config)

    def load_input_config(self, input_config: InputConfig):
        """Load an InputConfig form the active mapping input combination."""
        self.data_manager.load_input_config(input_config)

    def update_input_config(self, new_input_config: InputConfig):
        """Modify the active input configuration."""
        try:
            self.data_manager.update_input_config(new_input_config)
        except KeyError:
            # we need to synchronize the gui
            self.data_manager.publish_mapping()
            self.data_manager.publish_event()

    def remove_event(self):
        """Remove the active InputEvent from the active mapping event combination."""
        if (
            not self.data_manager.active_mapping
            or not self.data_manager.active_input_config
        ):
            return

        combination = list(self.data_manager.active_mapping.input_combination)
        combination.remove(self.data_manager.active_input_config)
        try:
            self.data_manager.update_mapping(
                input_combination=InputCombination(combination)
            )
            self.load_input_config(combination[0])
            self.save()
        except (KeyError, ValueError):
            # we need to synchronize the gui
            self.data_manager.publish_mapping()
            self.data_manager.publish_event()

    def set_event_as_analog(self, analog: bool):
        """Use the active event as an analog input."""
        assert self.data_manager.active_input_config is not None
        event = self.data_manager.active_input_config

        if event.type != EV_KEY:
            if analog:
                try:
                    self.data_manager.update_input_config(
                        event.modify(analog_threshold=0)
                    )
                    self.save()
                    return
                except KeyError:
                    pass
            else:
                try_values = {EV_REL: [1, -1], EV_ABS: [10, -10]}
                for value in try_values[event.type]:
                    try:
                        self.data_manager.update_input_config(
                            event.modify(analog_threshold=value)
                        )
                        self.save()
                        return
                    except KeyError:
                        pass

        # didn't update successfully
        # we need to synchronize the gui
        self.data_manager.publish_mapping()
        self.data_manager.publish_event()

    def load_groups(self):
        """Refresh the groups."""
        self.data_manager.refresh_groups()

    def load_group(self, group_key: str):
        """Load the group and then a preset of that group."""
        self.data_manager.load_group(group_key)
        self.load_preset(self.get_a_preset())

    def load_preset(self, name: str):
        """Load the preset."""
        self.data_manager.load_preset(name)
        # self.load_mapping(...) # not needed because we have on_preset_changed()

    def rename_preset(self, new_name: str):
        """Rename the active_preset."""
        if (
            not self.data_manager.active_preset
            or not new_name
            or new_name == self.data_manager.active_preset.name
        ):
            return

        new_name = PathUtils.sanitize_path_component(new_name)
        new_name = self.data_manager.get_available_preset_name(new_name)
        self.data_manager.rename_preset(new_name)

    def add_preset(self, name: str = DEFAULT_PRESET_NAME):
        """Create a new preset called `new preset n`, add it to the active_group."""
        name = self.data_manager.get_available_preset_name(name)
        try:
            self.data_manager.create_preset(name)
            self.data_manager.load_preset(name)
        except PermissionError as e:
            self.show_status(CTX_ERROR, _("Permission denied!"), str(e))

    def delete_preset(self):
        """Delete the active_preset from the disc."""

        def f(answer: bool):
            if answer:
                self.data_manager.delete_preset()
                self.data_manager.load_preset(self.get_a_preset())
                self.message_broker.publish(DoStackSwitch(1))

        if not self.data_manager.active_preset:
            return
        msg = (
            _('Are you sure you want to delete the preset "%s"?')
            % self.data_manager.active_preset.name
        )
        self.message_broker.publish(UserConfirmRequest(msg, f))

    def load_mapping(self, input_combination: InputCombination):
        """Load the mapping with the given input_combination form the active_preset."""
        self.data_manager.load_mapping(input_combination)
        self.load_input_config(input_combination[0])

    def update_mapping(self, **changes):
        """Update the active_mapping with the given keywords and values."""
        if "mapping_type" in changes.keys():
            if not (changes := self._change_mapping_type(changes)):
                # we need to synchronize the gui
                self.data_manager.publish_mapping()
                self.data_manager.publish_event()
                return

        self.data_manager.update_mapping(**changes)
        self.save()

    def create_mapping(self):
        """Create a new empty mapping in the active_preset."""
        try:
            self.data_manager.create_mapping()
        except KeyError:
            # there is already an empty mapping
            return

        self.data_manager.load_mapping(combination=InputCombination.empty_combination())
        self.data_manager.update_mapping(**MAPPING_DEFAULTS)

    def delete_mapping(self):
        """Remove the active_mapping form the active_preset."""

        def get_answer(answer: bool):
            if answer:
                self.data_manager.delete_mapping()
                self.save()

        if not self.data_manager.active_mapping:
            return
        self.message_broker.publish(
            UserConfirmRequest(
                _("Are you sure you want to delete this mapping?"),
                get_answer,
            )
        )

    def set_autoload(self, autoload: bool):
        """Set the autoload state for the active_preset and active_group."""
        self.data_manager.set_autoload(autoload)
        self.data_manager.refresh_service_config_path()

    def save(self):
        """Save all data to the disc."""
        try:
            self.data_manager.save()
        except PermissionError as e:
            self.show_status(CTX_ERROR, _("Permission denied!"), str(e))

    def start_key_recording(self):
        """Record the input of the active_group

        Updates the active_mapping.input_combination with the recorded events.
        """
        state = self.data_manager.get_state()
        if state == InjectorState.RUNNING or state == InjectorState.STARTING:
            self.data_manager.stop_combination_recording()
            self.message_broker.signal(MessageType.recording_finished)
            self.show_status(CTX_ERROR, _('Use "Stop" to stop before editing'))
            return

        logger.debug("Recording Keys")

        def on_recording_finished(_):
            self.message_broker.unsubscribe(on_recording_finished)
            self.message_broker.unsubscribe(self._on_combination_recorded)
            self.gui.connect_shortcuts()

        self.gui.disconnect_shortcuts()
        self.message_broker.subscribe(
            MessageType.combination_recorded,
            self._on_combination_recorded,
        )
        self.message_broker.subscribe(
            MessageType.recording_finished, on_recording_finished
        )
        self.data_manager.start_combination_recording()

    def stop_key_recording(self):
        """Stop recording the input."""
        logger.debug("Stopping Recording Keys")
        self.data_manager.stop_combination_recording()

    def start_injecting(self):
        """Inject the active_preset for the active_group."""
        if len(self.data_manager.active_preset) == 0:
            logger.error(_("Cannot apply empty preset file"))
            # also helpful for first time use
            self.show_status(CTX_ERROR, _("You need to add mappings first"))
            return

        if not self.button_left_warn:
            if self.data_manager.active_preset.dangerously_mapped_btn_left():
                self.show_status(
                    CTX_ERROR,
                    "This would disable your click button",
                    "Map a button to BTN_LEFT to avoid this.\n"
                    "To overwrite this warning, press apply again.",
                )
                self.button_left_warn = True
                return

        # todo: warn about unreleased keys
        self.button_left_warn = False
        self.message_broker.subscribe(
            MessageType.injector_state,
            self.show_injector_result,
        )
        self.show_status(CTX_APPLY, _("Starting injection..."))
        if not self.data_manager.start_injecting():
            self.message_broker.unsubscribe(self.show_injector_result)
            self.show_status(
                CTX_APPLY,
                _('Failed to apply preset "%s"') % self.data_manager.active_preset.name,
            )

    def show_injector_result(self, msg: InjectorStateMessage) -> None:
        """Show if the injection was successfully started."""
        self.message_broker.unsubscribe(self.show_injector_result)
        state = msg.state

        def running() -> None:
            assert self.data_manager.active_preset is not None
            msg = _('Applied preset "%s"') % self.data_manager.active_preset.name
            if self.data_manager.active_preset.dangerously_mapped_btn_left():
                msg += _(", CTRL + DEL to stop")
            self.show_status(CTX_APPLY, msg)
            logger.info(
                'Group "%s" is currently mapped', self.data_manager.active_group.key
            )

        def no_grab() -> None:
            assert self.data_manager.active_preset is not None
            msg = (
                _('Failed to apply preset "%s"') % self.data_manager.active_preset.name
            )
            tooltip = (
                "Maybe your preset doesn't contain anything that is sent by the "
                "device or another device is already grabbing it"
            )

            # InjectorState.NO_GRAB also happens when all mappings have validation
            # errors. In that case, we can show something more useful.
            validation_result = self._format_status_bar_validation_errors()
            if validation_result is not None:
                msg = f"{msg}. {validation_result[0]}"
                tooltip = validation_result[1]

            self.show_status(CTX_ERROR, msg, tooltip)

        assert self.data_manager.active_preset  # make mypy happy
        state_calls: Dict[InjectorState, Callable] = {
            InjectorState.RUNNING: running,
            InjectorState.ERROR: partial(
                self.show_status,
                CTX_ERROR,
                _('Error applying preset "%s"') % self.data_manager.active_preset.name,
            ),
            InjectorState.NO_GRAB: no_grab,
            InjectorState.UPGRADE_EVDEV: partial(
                self.show_status,
                CTX_ERROR,
                "Upgrade python-evdev",
                "Your python-evdev version is too old.",
            ),
        }

        if state in state_calls:
            state_calls[state]()

    def stop_injecting(self):
        """Stop injecting any preset for the active_group."""

        def show_result(msg: InjectorStateMessage):
            self.message_broker.unsubscribe(show_result)

            if not msg.inactive():
                # some speculation: there might be unexpected additional status messages
                # with a different state, or the status is wrong because something in
                # the long pipeline of status messages is broken.
                logger.error(
                    "Expected the injection to eventually stop, but got state %s",
                    msg.state,
                )
                return

            self.show_status(CTX_APPLY, _("Stopped the injection"))

        try:
            self.message_broker.subscribe(MessageType.injector_state, show_result)
            self.data_manager.stop_injecting()
        except DataManagementError:
            self.message_broker.unsubscribe(show_result)

    def show_status(
        self,
        ctx_id: int,
        msg: Optional[str] = None,
        tooltip: Optional[str] = None,
    ):
        """Send a status message to the ui to show it in the status-bar."""
        self.message_broker.publish(StatusData(ctx_id, msg, tooltip))

    def is_empty_mapping(self) -> bool:
        """Check if the active_mapping is empty."""
        return (
            self.data_manager.active_mapping == UIMapping(**MAPPING_DEFAULTS)
            or self.data_manager.active_mapping is None
        )

    def refresh_groups(self):
        """Reload the connected devices and send them as a groups message.

        Runs asynchronously.
        """
        self.data_manager.refresh_groups()

    def close(self):
        """Safely close the application."""
        logger.debug("Closing Application")
        self.save()
        self.message_broker.signal(MessageType.terminate)
        logger.debug("Quitting")
        Gtk.main_quit()

    def set_focus(self, component):
        """Focus the given component."""
        self.gui.window.set_focus(component)

    def _change_mapping_type(self, changes: Dict[str, Any]):
        """Query the user to update the mapping in order to change the mapping type."""
        mapping = self.data_manager.active_mapping

        if mapping is None:
            return changes

        if changes["mapping_type"] == mapping.mapping_type:
            return changes

        if changes["mapping_type"] == MappingType.ANALOG.value:
            msg = _("You are about to change the mapping to analog.")
            if mapping.output_symbol:
                msg += _('\nThis will remove "{}" ' "from the text input!").format(
                    mapping.output_symbol
                )

            if not [
                input_config
                for input_config in mapping.input_combination
                if input_config.defines_analog_input
            ]:
                # there is no analog input configured, let's try to autoconfigure it
                inputs: List[InputConfig] = list(mapping.input_combination)
                for i, input_config in enumerate(inputs):
                    if input_config.type in [EV_ABS, EV_REL]:
                        inputs[i] = input_config.modify(analog_threshold=0)
                        changes["input_combination"] = InputCombination(inputs)
                        msg += _(
                            '\nThe input "{}" will be used as analog input.'
                        ).format(input_config.description())
                        break
                else:
                    # not possible to autoconfigure inform the user
                    msg += _("\nYou need to record an analog input.")

            elif not mapping.output_symbol:
                return changes

            answer = None

            def get_answer(answer_: bool):
                nonlocal answer
                answer = answer_

            self.message_broker.publish(UserConfirmRequest(msg, get_answer))
            if answer:
                changes["output_symbol"] = None
                return changes
            else:
                return None

        if changes["mapping_type"] == MappingType.KEY_MACRO.value:
            try:
                analog_input = tuple(
                    filter(lambda i: i.defines_analog_input, mapping.input_combination)
                )[0]
            except IndexError:
                changes["output_type"] = None
                changes["output_code"] = None
                return changes

            answer = None

            def get_answer(answer_: bool):
                nonlocal answer
                answer = answer_

            self.message_broker.publish(
                UserConfirmRequest(
                    f"You are about to change the mapping to a Key or Macro mapping!\n"
                    f"Go to the advanced input configuration and set a "
                    f'"Trigger Threshold" for "{analog_input.description()}".',
                    get_answer,
                )
            )
            if answer:
                changes["output_type"] = None
                changes["output_code"] = None
                return changes
            else:
                return None

        return changes



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2023 sezanzeb <proxima@hip70890b.de>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Process that sends stuff to the GUI.

It should be started via input-remapper-control and pkexec.

GUIs should not run as root
https://wiki.archlinux.org/index.php/Running_GUI_applications_as_root

The service shouldn't do that even though it has root rights, because that
would enable key-loggers to just ask input-remapper for all user-input.

Instead, the ReaderService is used, which will be stopped when the gui closes.

Whereas for the reader-service to start a password is needed and it stops whe
the ui closes.

This uses the backend injection.event_reader and mapping_handlers to process all the
different input-events into simple on/off events and sends them to the gui.
"""

from __future__ import annotations

import asyncio
import logging
import multiprocessing
import os
import subprocess
import sys
import time
from collections import defaultdict
from typing import Set, List, Tuple

import evdev
from evdev.ecodes import EV_KEY, EV_ABS, EV_REL, REL_HWHEEL, REL_WHEEL

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import Mapping
from inputremapper.groups import _Groups, _Group
from inputremapper.injection.event_reader import EventReader
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.abs_to_btn_handler import AbsToBtnHandler
from inputremapper.injection.mapping_handlers.mapping_handler import (
    NotifyCallback,
    InputEventHandler,
    MappingHandler,
)
from inputremapper.injection.mapping_handlers.rel_to_btn_handler import RelToBtnHandler
from inputremapper.input_event import InputEvent, EventActions
from inputremapper.ipc.pipe import Pipe
from inputremapper.logging.logger import logger
from inputremapper.user import UserUtils
from inputremapper.utils import get_device_hash

# received by the reader-service
CMD_TERMINATE = "terminate"
CMD_STOP_READING = "stop-reading"
CMD_REFRESH_GROUPS = "refresh_groups"

# sent by the reader-service to the reader
MSG_GROUPS = "groups"
MSG_EVENT = "event"
MSG_STATUS = "status"


class ReaderService:
    """Service that only reads events and is supposed to run as root.

    Sends device information and keycodes to the GUI.

    Commands are either numbers for generic commands,
    or strings to start listening on a specific device.
    """

    # the speed threshold at which relative axis are considered moving
    # and will be sent as "pressed" to the frontend.
    # We want to allow some mouse movement before we record it as an input
    rel_xy_speed = defaultdict(lambda: 3)
    # wheel events usually don't produce values higher than 1
    rel_xy_speed[REL_WHEEL] = 1
    rel_xy_speed[REL_HWHEEL] = 1

    # Polkit won't ask for another password if the pid stays the same or something, and
    # if the previous request was no more than 5 minutes ago. see
    # https://unix.stackexchange.com/a/458260.
    # If the user does something after 6 minutes they will get a prompt already if the
    # reader timed out already, which sounds annoying. Instead, I'd rather have the
    # password prompt appear at most every 15 minutes.
    _maximum_lifetime: int = 60 * 15
    _timeout_tolerance: int = 60

    def __init__(self, groups: _Groups, global_uinputs: GlobalUInputs) -> None:
        """Construct the reader-service and initialize its communication pipes."""
        self._start_time = time.time()
        self.groups = groups
        self.global_uinputs = global_uinputs
        self._results_pipe = Pipe(self.get_pipe_paths()[0])
        self._commands_pipe = Pipe(self.get_pipe_paths()[1])
        self._pipe = multiprocessing.Pipe()

        self._tasks: Set[asyncio.Task] = set()
        self._stop_event = asyncio.Event()

        self._results_pipe.send({"type": MSG_STATUS, "message": "ready"})

    @staticmethod
    def get_pipe_paths() -> Tuple[str, str]:
        """Get the path where the pipe can be found."""
        return (
            f"/tmp/input-remapper-{UserUtils.home}/reader-results",
            f"/tmp/input-remapper-{UserUtils.home}/reader-commands",
        )

    @staticmethod
    def pipes_exist() -> bool:
        # Just checking for one of the 4 files (results, commands both read and write)
        # should be enough I guess.
        path = f"{ReaderService.get_pipe_paths()[0]}r"
        # Use os.path.exists, not lexists or islink, because broken links are bad.
        # New pipes and symlinks need to be made.
        return os.path.exists(path)

    @staticmethod
    def is_running() -> bool:
        """Check if the reader-service is running."""
        try:
            subprocess.check_output(["pgrep", "-f", "input-remapper-reader-service"])
        except subprocess.CalledProcessError:
            return False
        return True

    @staticmethod
    def pkexec_reader_service():
        """Start reader-service via pkexec to run in the background."""
        debug = " -d" if logger.level <= logging.DEBUG else ""
        cmd = f"pkexec input-remapper-control --command start-reader-service{debug}"

        logger.debug("Running `%s`", cmd)
        exit_code = os.system(cmd)

        if exit_code != 0:
            raise Exception(f"Failed to pkexec the reader-service, code {exit_code}")

    async def run(self):
        """Start doing stuff."""
        # the reader will check for new commands later, once it is running
        # it keeps running for one device or another.
        logger.debug("Discovering initial groups")
        self.groups.refresh()
        self._send_groups()
        await asyncio.gather(
            self._read_commands(),
            self._timeout(),
            self._stop_if_pipes_broken(),
        )

    def _send_groups(self):
        """Send the groups to the gui."""
        logger.debug("Sending groups")
        self._results_pipe.send({"type": MSG_GROUPS, "message": self.groups.dumps()})

    async def _timeout(self):
        """Stop automatically after some time."""
        # Prevents a permanent hole for key-loggers to exist, in case the gui crashes.
        # If the ReaderService stops even though the gui needs it, it needs to restart
        # it. This makes it also more comfortable to have debug mode running during
        # development, because it won't keep writing inputs containing passwords and
        # such to the terminal forever.

        await asyncio.sleep(self._maximum_lifetime)

        # if it is currently reading, wait a bit longer for the gui to complete
        # what it is doing.
        if self._is_reading():
            logger.debug("Waiting a bit longer for the gui to finish reading")

            for _ in range(self._timeout_tolerance):
                if not self._is_reading():
                    # once reading completes, it should terminate right away
                    break

                await asyncio.sleep(1)

        logger.debug("Maximum life-span reached, terminating")
        sys.exit(1)

    async def _read_commands(self):
        """Handle all unread commands.
        this will run until it receives CMD_TERMINATE
        """
        logger.debug("Waiting for commands")
        async for cmd in self._commands_pipe:
            logger.debug('Received command "%s"', cmd)

            if cmd == CMD_TERMINATE:
                await self._stop_reading()
                logger.debug("Terminating")
                sys.exit(0)

            if cmd == CMD_REFRESH_GROUPS:
                self.groups.refresh()
                self._send_groups()
                continue

            if cmd == CMD_STOP_READING:
                await self._stop_reading()
                continue

            group = self.groups.find(key=cmd)
            if group is None:
                # this will block for a bit maybe we want to do this async?
                self.groups.refresh()
                group = self.groups.find(key=cmd)

            if group is not None:
                await self._stop_reading()
                self._start_reading(group)
                continue

            logger.error('Received unknown command "%s"', cmd)

    async def _stop_if_pipes_broken(self):
        # The GUI probably exited, and failed to tell the reader-service to stop.
        # Pipes are owned by the GUI process, because the non-privileged GUI process
        # needs to be able to read them. Therefore, they are gone.
        while True:
            await asyncio.sleep(1)
            if not self.pipes_exist():
                await self._stop_reading()
                logger.debug("Pipes broken, exiting")
                sys.exit(13)

    def _is_reading(self) -> bool:
        """Check if the ReaderService is currently sending events to the GUI."""
        return len(self._tasks) > 0

    def _start_reading(self, group: _Group):
        """Find all devices of that group, filter interesting ones and send the events
        to the gui."""
        sources = []
        for path in group.paths:
            try:
                device = evdev.InputDevice(path)
            except (FileNotFoundError, OSError):
                logger.error('Could not find "%s"', path)
                return None

            capabilities = device.capabilities(absinfo=False)
            if (
                EV_KEY in capabilities
                or EV_ABS in capabilities
                or EV_REL in capabilities
            ):
                sources.append(device)

        context = self._create_event_pipeline(sources)
        # create the event reader and start it
        for device in sources:
            reader = EventReader(context, device, self._stop_event)
            self._tasks.add(asyncio.create_task(reader.run()))

    async def _stop_reading(self):
        """Stop the running event_reader."""
        self._stop_event.set()
        if self._tasks:
            await asyncio.gather(*self._tasks)
        self._tasks = set()
        self._stop_event.clear()

    def _create_event_pipeline(self, sources: List[evdev.InputDevice]) -> ContextDummy:
        """Create a custom event pipeline for each event code in the capabilities.

        Instead of sending the events to an uinput they will be sent to the frontend.
        """
        context_dummy = ContextDummy()
        # create a context for each source
        for device in sources:
            device_hash = get_device_hash(device)
            capabilities = device.capabilities(absinfo=False)

            for ev_code in capabilities.get(EV_KEY) or ():
                input_config = InputConfig(
                    type=EV_KEY, code=ev_code, origin_hash=device_hash
                )
                context_dummy.add_handler(
                    input_config, ForwardToUIHandler(self._results_pipe)
                )

            for ev_code in capabilities.get(EV_ABS) or ():
                # positive direction
                input_config = InputConfig(
                    type=EV_ABS,
                    code=ev_code,
                    analog_threshold=30,
                    origin_hash=device_hash,
                )
                mapping = Mapping(
                    input_combination=InputCombination([input_config]),
                    target_uinput="keyboard",
                    output_symbol="KEY_A",
                )
                handler: MappingHandler = AbsToBtnHandler(
                    InputCombination([input_config]),
                    mapping,
                    self.global_uinputs,
                )
                handler.set_sub_handler(ForwardToUIHandler(self._results_pipe))
                context_dummy.add_handler(input_config, handler)

                # negative direction
                input_config = input_config.modify(analog_threshold=-30)
                mapping = Mapping(
                    input_combination=InputCombination([input_config]),
                    target_uinput="keyboard",
                    output_symbol="KEY_A",
                )
                handler = AbsToBtnHandler(
                    InputCombination([input_config]),
                    mapping,
                    self.global_uinputs,
                )
                handler.set_sub_handler(ForwardToUIHandler(self._results_pipe))
                context_dummy.add_handler(input_config, handler)

            for ev_code in capabilities.get(EV_REL) or ():
                # positive direction
                input_config = InputConfig(
                    type=EV_REL,
                    code=ev_code,
                    analog_threshold=self.rel_xy_speed[ev_code],
                    origin_hash=device_hash,
                )
                mapping = Mapping(
                    input_combination=InputCombination([input_config]),
                    target_uinput="keyboard",
                    output_symbol="KEY_A",
                    release_timeout=0.3,
                    force_release_timeout=True,
                )
                handler = RelToBtnHandler(
                    InputCombination([input_config]),
                    mapping,
                    self.global_uinputs,
                )
                handler.set_sub_handler(ForwardToUIHandler(self._results_pipe))
                context_dummy.add_handler(input_config, handler)

                # negative direction
                input_config = input_config.modify(
                    analog_threshold=-self.rel_xy_speed[ev_code]
                )
                mapping = Mapping(
                    input_combination=InputCombination([input_config]),
                    target_uinput="keyboard",
                    output_symbol="KEY_A",
                    release_timeout=0.3,
                    force_release_timeout=True,
                )
                handler = RelToBtnHandler(
                    InputCombination([input_config]),
                    mapping,
                    self.global_uinputs,
                )
                handler.set_sub_handler(ForwardToUIHandler(self._results_pipe))
                context_dummy.add_handler(input_config, handler)

        return context_dummy


class ForwardDummy:
    @staticmethod
    def write(*_):
        pass


class ContextDummy:
    """Used for the reader so that no events are actually written to any uinput."""

    def __init__(self):
        self.listeners = set()
        self._notify_callbacks = defaultdict(list)
        self.forward_dummy = ForwardDummy()

    def add_handler(self, input_config: InputConfig, handler: InputEventHandler):
        self._notify_callbacks[input_config.input_match_hash].append(handler.notify)

    def get_notify_callbacks(self, input_event: InputEvent) -> List[NotifyCallback]:
        return self._notify_callbacks[input_event.input_match_hash]

    def reset(self):
        pass

    def get_forward_uinput(self, origin_hash) -> evdev.UInput:
        """Don't actually write anything."""
        return self.forward_dummy


class ForwardToUIHandler:
    """Implements the InputEventHandler protocol. Sends all events into the pipe."""

    def __init__(self, pipe: Pipe):
        self.pipe = pipe
        self._last_event = InputEvent.from_tuple((99, 99, 99))

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        """Filter duplicates and send into the pipe."""
        if event != self._last_event:
            self._last_event = event
            if EventActions.negative_trigger in event.actions:
                event = event.modify(value=-1)

            logger.debug("Sending to %s frontend", event)
            self.pipe.send(
                {
                    "type": MSG_EVENT,
                    "message": {
                        "sec": event.sec,
                        "usec": event.usec,
                        "type": event.type,
                        "code": event.code,
                        "value": event.value,
                        "origin_hash": event.origin_hash,
                    },
                }
            )
        return True

    def reset(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.task import Task


class KeyTask(Task):
    """Write the symbol."""

    argument_configs = [
        ArgumentConfig(
            name="symbol",
            position=0,
            types=[str],
            is_symbol=True,
        )
    ]

    async def run(self, callback) -> None:
        symbol = self.get_argument("symbol").get_value()
        code = keyboard_layout.get(symbol)

        callback(EV_KEY, code, 1)
        await self.keycode_pause()
        callback(EV_KEY, code, 0)
        await self.keycode_pause()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Testing the input-remapper-control command"""

import collections
import os
import time
import unittest
from unittest.mock import patch, MagicMock

from inputremapper.bin.input_remapper_control import InputRemapperControlBin
from inputremapper.configs.global_config import GlobalConfig
from inputremapper.configs.migrations import Migrations
from inputremapper.configs.paths import PathUtils
from inputremapper.configs.preset import Preset
from inputremapper.daemon import Daemon
from inputremapper.groups import groups
from inputremapper.injection.global_uinputs import GlobalUInputs, FrontendUInput
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from tests.lib.test_setup import test_setup
from tests.lib.tmp import tmp

options = collections.namedtuple(
    "options",
    ["command", "config_dir", "preset", "device", "list_devices", "key_names", "debug"],
)


def remove_timeout_from_calls(method: str):
    # Remove the timeout argument, which is used by dasbus but not actually
    # passed to the Daemon class methods. Since we create the Daemon object directly
    # for this test it would raise an exception otherwise, for example:
    # TypeError: Daemon.autoload_single() got an unexpected keyword argument 'timeout'
    def decorator(func):
        def wrapped(*args, **kwargs):
            original_method = getattr(Daemon, method)
            with patch.object(
                Daemon,
                method,
                lambda *args, timeout=0, **kwargs: original_method(*args, **kwargs),
            ):
                return func(*args, **kwargs)

        return wrapped

    return decorator


@test_setup
class TestControl(unittest.TestCase):
    def setUp(self):
        self.global_config = GlobalConfig()
        self.global_uinputs = GlobalUInputs(FrontendUInput)
        self.migrations = Migrations(self.global_uinputs)
        self.mapping_parser = MappingParser(self.global_uinputs)
        self.input_remapper_control = InputRemapperControlBin(
            self.global_config, self.migrations
        )

    @remove_timeout_from_calls("autoload")
    @remove_timeout_from_calls("autoload_single")
    def test_autoload(self):
        device_keys = ["Foo Device 2", "Bar Device"]
        groups_ = [groups.find(key=key) for key in device_keys]
        presets = ["bar0", "bar", "bar2"]
        paths = [
            PathUtils.get_preset_path(groups_[0].name, presets[0]),
            PathUtils.get_preset_path(groups_[1].name, presets[1]),
            PathUtils.get_preset_path(groups_[1].name, presets[2]),
        ]

        Preset(paths[0]).save()
        Preset(paths[1]).save()
        Preset(paths[2]).save()

        daemon = Daemon(self.global_config, self.global_uinputs, self.mapping_parser)

        self.input_remapper_control.set_daemon(daemon)

        start_history = []
        stop_counter = 0

        # using an actual injector is not within the scope of this test
        class Injector:
            def stop_injecting(self, *args, **kwargs):
                nonlocal stop_counter
                stop_counter += 1

        def start_injecting(device: str, preset: str):
            print(f'\033[90mstart_injecting "{device}" "{preset}"\033[0m')
            start_history.append((device, preset))
            daemon.injectors[device] = Injector()

        patch.object(daemon, "start_injecting", start_injecting).start()

        self.global_config.set_autoload_preset(groups_[0].key, presets[0])
        self.global_config.set_autoload_preset(groups_[1].key, presets[1])

        self.input_remapper_control.communicate(
            command="autoload",
            config_dir=None,
            preset=None,
            device=None,
        )

        self.assertEqual(len(start_history), 2)
        self.assertEqual(start_history[0], (groups_[0].key, presets[0]))
        self.assertEqual(start_history[1], (groups_[1].key, presets[1]))
        self.assertIn(groups_[0].key, daemon.injectors)
        self.assertIn(groups_[1].key, daemon.injectors)
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[0].key, presets[0])
        )
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[1].key, presets[1])
        )

        # calling autoload again doesn't load redundantly
        self.input_remapper_control.communicate(
            command="autoload",
            config_dir=None,
            preset=None,
            device=None,
        )
        self.assertEqual(len(start_history), 2)
        self.assertEqual(stop_counter, 0)
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[0].key, presets[0])
        )
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[1].key, presets[1])
        )

        # unless the injection in question ist stopped
        self.input_remapper_control.communicate(
            command="stop",
            config_dir=None,
            preset=None,
            device=groups_[0].key,
        )
        self.assertEqual(stop_counter, 1)
        self.assertTrue(
            daemon.autoload_history.may_autoload(groups_[0].key, presets[0])
        )
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[1].key, presets[1])
        )
        self.input_remapper_control.communicate(
            command="autoload",
            config_dir=None,
            preset=None,
            device=None,
        )
        self.assertEqual(len(start_history), 3)
        self.assertEqual(start_history[2], (groups_[0].key, presets[0]))
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[0].key, presets[0])
        )
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[1].key, presets[1])
        )

        # if a device name is passed, will only start injecting for that one
        self.input_remapper_control.communicate(
            command="stop-all",
            config_dir=None,
            preset=None,
            device=None,
        )
        self.assertTrue(
            daemon.autoload_history.may_autoload(groups_[0].key, presets[0])
        )
        self.assertTrue(
            daemon.autoload_history.may_autoload(groups_[1].key, presets[1])
        )
        self.assertEqual(stop_counter, 3)
        self.global_config.set_autoload_preset(groups_[1].key, presets[2])
        self.input_remapper_control.communicate(
            command="autoload",
            config_dir=None,
            preset=None,
            device=groups_[1].key,
        )
        self.assertEqual(len(start_history), 4)
        self.assertEqual(start_history[3], (groups_[1].key, presets[2]))
        self.assertTrue(
            daemon.autoload_history.may_autoload(groups_[0].key, presets[0])
        )
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[1].key, presets[2])
        )

        # autoloading for the same device again redundantly will not autoload
        # again
        self.input_remapper_control.communicate(
            command="autoload",
            config_dir=None,
            preset=None,
            device=groups_[1].key,
        )
        self.assertEqual(len(start_history), 4)
        self.assertEqual(stop_counter, 3)
        self.assertFalse(
            daemon.autoload_history.may_autoload(groups_[1].key, presets[2])
        )

        # any other arbitrary preset may be autoloaded
        self.assertTrue(daemon.autoload_history.may_autoload(groups_[1].key, "quuuux"))

        # after 15 seconds it may be autoloaded again
        daemon.autoload_history._autoload_history[groups_[1].key] = (
            time.time() - 16,
            presets[2],
        )
        self.assertTrue(
            daemon.autoload_history.may_autoload(groups_[1].key, presets[2])
        )

    @remove_timeout_from_calls("autoload")
    def test_autoload_other_path(self):
        device_names = ["Foo Device", "Bar Device"]
        groups_ = [groups.find(name=name) for name in device_names]
        presets = ["bar123", "bar2"]
        config_dir = os.path.join(tmp, "qux", "quux")
        paths = [
            os.path.join(config_dir, "presets", device_names[0], presets[0] + ".json"),
            os.path.join(config_dir, "presets", device_names[1], presets[1] + ".json"),
        ]

        Preset(paths[0]).save()
        Preset(paths[1]).save()

        daemon = Daemon(self.global_config, self.global_uinputs, self.mapping_parser)
        self.input_remapper_control.set_daemon(daemon)

        start_history = []
        daemon.start_injecting = lambda *args: start_history.append(args)

        self.global_config.path = os.path.join(config_dir, "config.json")
        self.global_config.load_config()
        self.global_config.set_autoload_preset(device_names[0], presets[0])
        self.global_config.set_autoload_preset(device_names[1], presets[1])

        self.input_remapper_control.communicate(
            command="autoload",
            config_dir=config_dir,
            preset=None,
            device=None,
        )

        self.assertEqual(len(start_history), 2)
        self.assertEqual(start_history[0], (groups_[0].key, presets[0]))
        self.assertEqual(start_history[1], (groups_[1].key, presets[1]))

    def test_start_stop(self):
        group = groups.find(key="Foo Device 2")
        preset = "preset9"

        daemon = Daemon(self.global_config, self.global_uinputs, self.mapping_parser)
        self.input_remapper_control.set_daemon(daemon)

        start_history = []
        stop_history = []
        stop_all_history = []
        daemon.start_injecting = lambda *args: start_history.append(args)
        daemon.stop_injecting = lambda *args: stop_history.append(args)
        daemon.stop_all = lambda *args: stop_all_history.append(args)

        self.input_remapper_control.communicate(
            command="start",
            config_dir=None,
            preset=preset,
            device=group.paths[0],
        )
        self.assertEqual(len(start_history), 1)
        self.assertEqual(start_history[0], (group.key, preset))

        self.input_remapper_control.communicate(
            command="stop",
            config_dir=None,
            preset=None,
            device=group.paths[1],
        )
        self.assertEqual(len(stop_history), 1)
        # provided any of the groups paths as --device argument, figures out
        # the correct group.key to use here
        self.assertEqual(stop_history[0], (group.key,))

        self.input_remapper_control.communicate(
            command="stop-all",
            config_dir=None,
            preset=None,
            device=None,
        )
        self.assertEqual(len(stop_all_history), 1)
        self.assertEqual(stop_all_history[0], ())

    @patch.object(Daemon, "quit")
    def test_quit(self, quit_mock: MagicMock) -> None:
        group = groups.find(key="Foo Device 2")
        assert group is not None
        preset = "preset9"

        daemon = Daemon(self.global_config, self.global_uinputs, self.mapping_parser)
        self.input_remapper_control.set_daemon(daemon)

        self.input_remapper_control.communicate(
            command="quit",
            config_dir=None,
            preset=preset,
            device=group.paths[0],
        )

        quit_mock.assert_called_once()

    def test_config_not_found(self):
        key = "Foo Device 2"
        path = "~/a/preset.json"
        config_dir = "/foo/bar"

        daemon = Daemon(self.global_config, self.global_uinputs, self.mapping_parser)
        self.input_remapper_control.set_daemon(daemon)

        start_history = []
        stop_history = []
        daemon.start_injecting = lambda *args: start_history.append(args)
        daemon.stop_injecting = lambda *args: stop_history.append(args)

        self.assertRaises(
            SystemExit,
            lambda: self.input_remapper_control.communicate(
                command="start",
                config_dir=config_dir,
                preset=path,
                device=key,
            ),
        )

        self.assertRaises(
            SystemExit,
            lambda: self.input_remapper_control.communicate(
                command="stop",
                config_dir=config_dir,
                preset=None,
                device=key,
            ),
        )

    def test_autoload_config_dir(self):
        daemon = Daemon(self.global_config, self.global_uinputs, self.mapping_parser)

        path = os.path.join(tmp, "foo")
        os.makedirs(path)
        with open(os.path.join(path, "config.json"), "w") as file:
            file.write('{"autoload":{"foo": "bar"}}')

        self.assertIsNone(self.global_config.get_autoload_preset("foo"))
        daemon.set_config_dir(path)
        # since daemon and this test share the same memory, the global_config
        # object that this test can access will be modified
        self.assertEqual(self.global_config.get_autoload_preset("foo"), "bar")

        # passing a path that doesn't exist or a path that doesn't contain
        # a config.json file won't do anything
        os.makedirs(os.path.join(tmp, "bar"))
        daemon.set_config_dir(os.path.join(tmp, "bar"))
        self.assertEqual(self.global_config.get_autoload_preset("foo"), "bar")
        daemon.set_config_dir(os.path.join(tmp, "qux"))
        self.assertEqual(self.global_config.get_autoload_preset("foo"), "bar")

    def test_internals_reader(self):
        with patch.object(os, "system") as os_system_patch:
            self.input_remapper_control.internals("start-reader-service", False)
            os_system_patch.assert_called_once()
            self.assertIn(
                "input-remapper-reader-service", os_system_patch.call_args.args[0]
            )
            self.assertNotIn("-d", os_system_patch.call_args.args[0])

    def test_internals_daemon(self):
        with patch.object(os, "system") as os_system_patch:
            self.input_remapper_control.internals("start-daemon", True)
            os_system_patch.assert_called_once()
            self.assertIn("input-remapper-service", os_system_patch.call_args.args[0])
            self.assertIn("-d", os_system_patch.call_args.args[0])


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import os.path
import re
import traceback
from collections import defaultdict, deque
from typing import (
    Callable,
    Dict,
    Set,
    Protocol,
    Tuple,
    Deque,
    Any,
)

from inputremapper.gui.messages.message_types import MessageType
from inputremapper.logging.logger import logger


class Message(Protocol):
    """The protocol any message must follow to be sent with the MessageBroker."""

    @property
    def message_type(self) -> MessageType: ...


# useful type aliases
MessageListener = Callable[[Any], None]


class MessageBroker:
    shorten_path = re.compile(r"inputremapper/")

    def __init__(self):
        self._listeners: Dict[MessageType, Set[MessageListener]] = defaultdict(set)
        self._messages: Deque[Tuple[Message, str, int]] = deque()
        self._publishing = False

    def publish(self, data: Message):
        """Schedule a massage to be sent.
        The message will be sent after all currently pending messages are sent."""
        self._messages.append((data, *self.get_caller()))
        self._publish_all()

    def signal(self, signal: MessageType):
        """Send a signal without any data payload."""
        # This is different from calling self.publish because self.get_caller()
        # looks back at the current stack 3 frames
        self._messages.append((Signal(signal), *self.get_caller()))
        self._publish_all()

    def _publish(self, data: Message, file: str, line: int):
        logger.debug(
            "from %s:%d: Signal=%s: %s", file, line, data.message_type.name, data
        )
        for listener in self._listeners[data.message_type].copy():
            listener(data)

    def _publish_all(self):
        """Send all scheduled messages in order."""
        if self._publishing:
            # don't run this twice, so we not mess up the order
            return

        self._publishing = True
        try:
            while self._messages:
                self._publish(*self._messages.popleft())
        finally:
            self._publishing = False

    def subscribe(self, massage_type: MessageType, listener: MessageListener):
        """Attach a listener to an event."""
        logger.debug("adding new Listener for %s: %s", massage_type, listener)
        self._listeners[massage_type].add(listener)
        return self

    @staticmethod
    def get_caller(position: int = 3) -> Tuple[str, int]:
        """Extract a file and line from current stack and format for logging."""
        tb = traceback.extract_stack(limit=position)[0]
        return os.path.basename(tb.filename), tb.lineno or 0

    def unsubscribe(self, listener: MessageListener) -> None:
        for listeners in self._listeners.values():
            try:
                listeners.remove(listener)
            except KeyError:
                pass


class Signal:
    """Send a Message without any associated data over the MassageBus."""

    def __init__(self, message_type: MessageType):
        self.message_type: MessageType = message_type

    def __str__(self):
        return f"Signal: {self.message_type}"

    def __eq__(self, other: Any):
        return type(self) is type(other) and self.message_type == other.message_type
